#include <sta/rtos/thread.hpp>
#include <sta/debug/assert.hpp>

#define IS_THREAD_SYS_FLAGS(flags)	( ( flags & STA_RTOS_THREAD_FLAGS_VALID_SYS_BITS ) == flags )
#define IS_THREAD_USER_FLAGS(flags)	( ( flags & STA_RTOS_THREAD_FLAGS_VALID_USER_BITS ) == flags )


namespace sta
{
	RtosThread::RtosThread(const char* name, osPriority_t prio, uint32_t stack_size /* = 0 */, uint32_t cb_size /* = 0 */)
		: instance_{NULL},
		  attribs_{ .name = name, .cb_size = cb_size, .stack_size = stack_size, .priority = prio }
	{}

	osThreadId_t RtosThread::getInstance()
	{
		return instance_;
	}

	const char* RtosThread::getName() const
	{
		return attribs_.name;
	}

	void RtosThread::start()
	{
		// Check if there is no instance that is currently running.
		STA_ASSERT(instance_ == NULL);

		instance_ = osThreadNew(entry_point, this, &attribs_);
		STA_ASSERT(instance_ != NULL);

		// Send a thread start signal.
		sysNotify(STA_RTOS_THREAD_FLAG_START);
	}

	void RtosThread::entry_point(void* arg)
	{
		STA_ASSERT(arg != nullptr);

		RtosThread* instance = reinterpret_cast<RtosThread*>(arg) ;
		instance->loop();
	}

	void RtosThread::sleep(uint32_t ticks)
	{
		// Make sure that the calling thread is this thread. If not, the
		// developer very likely has made a mistake. 
		STA_ASSERT(osThreadGetId() == instance_);

		osDelay(ticks);
	}

	void RtosThread::notify(uint32_t flags)
	{
		STA_ASSERT(instance_ != NULL);
		STA_ASSERT_MSG(IS_THREAD_USER_FLAGS(flags), "Only user flags allowed");

		osThreadFlagsSet(instance_, flags);
	}

	void RtosThread::sysNotify(uint32_t flags)
	{
		STA_ASSERT(instance_ != NULL);
		STA_ASSERT_MSG(IS_THREAD_SYS_FLAGS(flags), "Only system flags allowed");

		uint32_t rslt = osThreadFlagsSet(instance_, flags);

		STA_ASSERT_MSG(rslt != osFlagsErrorUnknown, "Unknown thread flag error.");
		STA_ASSERT_MSG(rslt != osFlagsErrorParameter, "Parameter thread_id is not a valid thread or flags has highest bit set.");
		STA_ASSERT_MSG(rslt != osFlagsErrorResource, "The thread is in invalid state.");
	}

	uint32_t RtosThread::wait(uint32_t flags)
	{
		// Make sure that the calling thread is this thread. If not, the
		// developer very likely has made a mistake. 
		STA_ASSERT(osThreadGetId() == instance_);

		return osThreadFlagsWait(flags, osFlagsWaitAny, osWaitForever);
	}

	uint32_t RtosThread::clear(uint32_t flags)
	{
		// Make sure that the calling thread is this thread. If not, the
		// developer very likely has made a mistake. 
		STA_ASSERT(osThreadGetId() == instance_);

		uint32_t setFlags = osThreadFlagsClear(flags);

		STA_ASSERT(setFlags != (uint32_t)osError);

		return setFlags;
	}

	uint32_t RtosThread::getFlags()
	{
		// Make sure that the calling thread is this thread. If not, the
		// developer very likely has made a mistake. 
		STA_ASSERT(osThreadGetId() == instance_);

		return osThreadFlagsGet();
	}

	void RtosThread::kill()
	{
		STA_ASSERT(instance_ != NULL);
		osThreadTerminate(instance_);
		instance_ = NULL;
	}
} // namespace sta
